Package: DelayedArray
Title: Delayed operations on array-like objects
Description: Wrapping an array-like object (typically an on-disk object) in
	a DelayedArray object allows one to perform common array operations
	on it without loading the object in memory. In order to reduce memory
	usage and optimize performance, operations on the object are either
	delayed or executed using a block processing mechanism. Note that this
	also works on in-memory array-like objects like DataFrame objects
	(typically with Rle columns), Matrix objects, and ordinary arrays and
	data frames.
Version: 0.6.5
Encoding: UTF-8
Author: Hervé Pagès
Maintainer: Hervé Pagès <hpages@fredhutch.org>
biocViews: Infrastructure, DataRepresentation, Annotation,
        GenomeAnnotation
Depends: R (>= 3.4), methods, stats4, matrixStats, BiocGenerics (>=
        0.25.1), S4Vectors (>= 0.17.43), IRanges (>= 2.11.17),
        BiocParallel
Imports: stats
Suggests: Matrix, HDF5Array, genefilter, SummarizedExperiment, airway,
        pryr, knitr, BiocStyle, RUnit
License: Artistic-2.0
VignetteBuilder: knitr
Collate: utils.R bind-arrays.R Array-class.R ArrayGrid-class.R
        extract_array.R show-utils.R DelayedOp-class.R showtree.R
        DelayedArray-class.R realize.R block_processing.R
        DelayedArray-utils.R DelayedMatrix-utils.R DelayedArray-stats.R
        DelayedMatrix-stats.R RleArray-class.R zzz.R
git_url: https://git.bioconductor.org/packages/DelayedArray
git_branch: RELEASE_3_7
git_last_commit: 7d1cb64
git_last_commit_date: 2018-08-14
Date/Publication: 2018-08-15
NeedsCompilation: no
Packaged: 2018-08-16 01:25:27 UTC; biocbuild
