[<img src="https://www.bioconductor.org/images/logo/jpg/bioconductor_logo_rgb.jpg" width="200" align="right"/>](https://bioconductor.org/)

_DelayedArray_ is an R/Bioconductor package that provides a unified framework for working transparently with on-disk and in-memory array-like datasets.

See https://bioconductor.org/packages/DelayedArray for more information including how to install the release version of the package (please refrain from installing directly from GitHub).

